# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class SessionsQuery(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'TextQuery',
        'paging': 'PageQuery',
        'state_filter': 'list[SessionState]',
        'creator_id': 'str'
    }

    attribute_map = {
        'name': 'name',
        'paging': 'paging',
        'state_filter': 'state_filter',
        'creator_id': 'creator_id'
    }

    def __init__(self, name=None, paging=None, state_filter=[], creator_id=None, local_vars_configuration=None):  # noqa: E501
        """SessionsQuery - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._paging = None
        self._state_filter = None
        self._creator_id = None
        self.discriminator = None

        if name is not None:
            self.name = name
        if paging is not None:
            self.paging = paging
        if state_filter is not None:
            self.state_filter = state_filter
        if creator_id is not None:
            self.creator_id = creator_id

    @property
    def name(self):
        """Gets the name of this SessionsQuery.  # noqa: E501

        Filters Cluster by name. If this field is absent, no filtering is done.  # noqa: E501

        :return: The name of this SessionsQuery.  # noqa: E501
        :rtype: TextQuery
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this SessionsQuery.

        Filters Cluster by name. If this field is absent, no filtering is done.  # noqa: E501

        :param name: The name of this SessionsQuery.  # noqa: E501
        :type: TextQuery
        """

        self._name = name

    @property
    def paging(self):
        """Gets the paging of this SessionsQuery.  # noqa: E501

        Pagination information.  # noqa: E501

        :return: The paging of this SessionsQuery.  # noqa: E501
        :rtype: PageQuery
        """
        return self._paging

    @paging.setter
    def paging(self, paging):
        """Sets the paging of this SessionsQuery.

        Pagination information.  # noqa: E501

        :param paging: The paging of this SessionsQuery.  # noqa: E501
        :type: PageQuery
        """

        self._paging = paging

    @property
    def state_filter(self):
        """Gets the state_filter of this SessionsQuery.  # noqa: E501

        Filter Cluster by Session State. If this field is an empty set, no filtering is done.  # noqa: E501

        :return: The state_filter of this SessionsQuery.  # noqa: E501
        :rtype: list[SessionState]
        """
        return self._state_filter

    @state_filter.setter
    def state_filter(self, state_filter):
        """Sets the state_filter of this SessionsQuery.

        Filter Cluster by Session State. If this field is an empty set, no filtering is done.  # noqa: E501

        :param state_filter: The state_filter of this SessionsQuery.  # noqa: E501
        :type: list[SessionState]
        """

        self._state_filter = state_filter

    @property
    def creator_id(self):
        """Gets the creator_id of this SessionsQuery.  # noqa: E501

        Filters Jobs by creator_id. If this field is absent, no filtering is done.  # noqa: E501

        :return: The creator_id of this SessionsQuery.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this SessionsQuery.

        Filters Jobs by creator_id. If this field is absent, no filtering is done.  # noqa: E501

        :param creator_id: The creator_id of this SessionsQuery.  # noqa: E501
        :type: str
        """

        self._creator_id = creator_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SessionsQuery):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SessionsQuery):
            return True

        return self.to_dict() != other.to_dict()
