# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class LogsOutput(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'lines': 'str',
        'num_lines': 'int',
        'start_line': 'int',
        'finished': 'bool'
    }

    attribute_map = {
        'lines': 'lines',
        'num_lines': 'num_lines',
        'start_line': 'start_line',
        'finished': 'finished'
    }

    def __init__(self, lines=None, num_lines=None, start_line=None, finished=None, local_vars_configuration=None):  # noqa: E501
        """LogsOutput - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._lines = None
        self._num_lines = None
        self._start_line = None
        self._finished = None
        self.discriminator = None

        self.lines = lines
        self.num_lines = num_lines
        self.start_line = start_line
        self.finished = finished

    @property
    def lines(self):
        """Gets the lines of this LogsOutput.  # noqa: E501


        :return: The lines of this LogsOutput.  # noqa: E501
        :rtype: str
        """
        return self._lines

    @lines.setter
    def lines(self, lines):
        """Sets the lines of this LogsOutput.


        :param lines: The lines of this LogsOutput.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and lines is None:  # noqa: E501
            raise ValueError("Invalid value for `lines`, must not be `None`")  # noqa: E501

        self._lines = lines

    @property
    def num_lines(self):
        """Gets the num_lines of this LogsOutput.  # noqa: E501


        :return: The num_lines of this LogsOutput.  # noqa: E501
        :rtype: int
        """
        return self._num_lines

    @num_lines.setter
    def num_lines(self, num_lines):
        """Sets the num_lines of this LogsOutput.


        :param num_lines: The num_lines of this LogsOutput.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_lines is None:  # noqa: E501
            raise ValueError("Invalid value for `num_lines`, must not be `None`")  # noqa: E501

        self._num_lines = num_lines

    @property
    def start_line(self):
        """Gets the start_line of this LogsOutput.  # noqa: E501


        :return: The start_line of this LogsOutput.  # noqa: E501
        :rtype: int
        """
        return self._start_line

    @start_line.setter
    def start_line(self, start_line):
        """Sets the start_line of this LogsOutput.


        :param start_line: The start_line of this LogsOutput.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and start_line is None:  # noqa: E501
            raise ValueError("Invalid value for `start_line`, must not be `None`")  # noqa: E501

        self._start_line = start_line

    @property
    def finished(self):
        """Gets the finished of this LogsOutput.  # noqa: E501


        :return: The finished of this LogsOutput.  # noqa: E501
        :rtype: bool
        """
        return self._finished

    @finished.setter
    def finished(self, finished):
        """Sets the finished of this LogsOutput.


        :param finished: The finished of this LogsOutput.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and finished is None:  # noqa: E501
            raise ValueError("Invalid value for `finished`, must not be `None`")  # noqa: E501

        self._finished = finished

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LogsOutput):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LogsOutput):
            return True

        return self.to_dict() != other.to_dict()
